package laboratoire4.strategies;

import laboratoire4.Action;
import laboratoire4.IPawn;
import laboratoire4.game.Game;
import laboratoire4.game.GameUtils;
import laboratoire4.game.SimulatedGame;
import laboratoire4.pawns.Pawn;
import laboratoire4.pawns.PawnMovement;
import laboratoire4.pawns.PawnUtils;
import laboratoire4.pawns.SimulatedPawn;

import java.util.Collection;
import java.util.stream.Collectors;

public abstract class MiniMaxStrategy implements Strategy {
    private static final int MAX_DEPTH = 6;
    private static final long MAX_TIME_MS = 4800;   // Moins de 5 secondes, car le tour n'est pas complètement fini

    protected SimulatedGame game;
    protected long startTimeMs;

    protected MiniMaxStrategy(Game game) {
        this.game = new SimulatedGame(game.getBoard(), game.getPlayer());
    }

    @Override
    public EvaluationResult getNextMove() {
        return miniMax();
    }

    private EvaluationResult miniMax() {
        EvaluationResult maxResult = null;
        int maxScore = Integer.MIN_VALUE;
        startTimeMs = System.currentTimeMillis();

        for (Action<IPawn> action : getActions(true)) {
            int score = min(0, Integer.MIN_VALUE, Integer.MAX_VALUE);

            if (maxResult == null || score > maxScore) {
                IPawn pawn = action.getPawn();
                maxResult = new EvaluationResult(pawn.getRow(), pawn.getCol(), action.getMovement());
            }
        }

        return maxResult;
    }

    private int max(int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluate();
        }

        if (System.currentTimeMillis() - startTimeMs > MAX_TIME_MS) {
            return Integer.MIN_VALUE;
        }

        int maxScore = Integer.MIN_VALUE;

        for (Action<IPawn> action : getActions(true)) {
            IPawn pawn = action.getPawn();
            PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = min(depth + 1, Math.max(alpha, maxScore), beta);
            game.revertMove();

            if (score > maxScore) {
                maxScore = score;
            }

            if (maxScore >= beta) {
                return maxScore;
            }
        }

        return maxScore;
    }

    private int min(int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluate();
        }

        if (System.currentTimeMillis() - startTimeMs > MAX_TIME_MS) {
            return Integer.MAX_VALUE;
        }

        int minScore = Integer.MAX_VALUE;

        for (Action<IPawn> action : getActions(false)) {
            IPawn pawn = action.getPawn();
            PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = max(depth + 1, alpha, Math.min(beta, minScore));
            game.revertMove();

            if (score < minScore) {
                minScore = score;
            }

            if (minScore <= alpha) {
                return minScore;
            }
        }

        return minScore;
    }

    protected Collection<SimulatedPawn> getMaxPawns() {
        return GameUtils.getMaxPawns(game.getBoard(), game.getPlayer());
    }

    protected Collection<SimulatedPawn> getMinPawns() {
        return GameUtils.getMinPawns(game.getBoard(), game.getPlayer());
    }

    protected Collection<Action<IPawn>> getActions(boolean max) {
        return PawnUtils.getActions(game, max);
    }

    private int evaluate() {
        int score = evaluateSimulation();

        Collection<SimulatedPawn> maxPawns = getMaxPawns();

        if (maxPawns.stream().noneMatch(IPawn::isPusher)) {
            return Integer.MIN_VALUE;
        }

        if (getMinPawns().stream().anyMatch(p -> p.getRow() == p.getPlayer().getGoal())) {
            return Integer.MIN_VALUE;
        }

        for (IPawn pawn : maxPawns) {
            if (pawn.getRow() == pawn.getPlayer().getGoal()) {
                return Integer.MAX_VALUE;
            }

            if (PawnUtils.canBeCaptured(game, pawn)) {
                // On peut être capturé
                score -= 10;
            }

            for (PawnMovement movement : PawnMovement.values()) {
                if (pawn.isMoveValid(game.getBoard(), movement) && !PawnUtils.canBeCaptured(game, pawn.getRow() + pawn.getDirection(), pawn.getCol() + movement.getMove(), pawn.getPlayer())) {
                    // On ne peut pas être capturé en faisant ce mouvement
                    score += 5;
                }
            }
        }

        // region Lignes ennemies
        for (IPawn pawn : getMinPawns()) {
            int row = pawn.getRow();
            int col = pawn.getCol();

            if (row > 0 && PawnUtils.areSamePlayers(pawn, game.getBoard()[row - 1][col])) {
                score -= 5;
            }
            if (row < 0 && PawnUtils.areSamePlayers(pawn, game.getBoard()[row + 1][col])) {
                score -= 5;
            }
        }
        // endregion

        return score;
    }

    protected abstract int evaluateSimulation();
}
